<?php
/**
 * SeekQuarry/Yioop --
 * Open Source Pure PHP Search Engine, Crawler, and Indexer
 *
 * Copyright (C) 2009 - 2019  Chris Pollett chris@pollett.org
 *
 * LICENSE:
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * END LICENSE
 *
 * @author Mangesh Dahale mangeshadahale@gmail.com
 * @license https://www.gnu.org/licenses/ GPL3
 * @link https://www.seekquarry.com/
 * @copyright 2009 - 2019
 * @filesource
 */
namespace seekquarry\yioop\library\summarizers;

use seekquarry\yioop\configs as C;
use seekquarry\yioop\library as L;
use seekquarry\yioop\library\CrawlConstants;
use seekquarry\yioop\library\PhraseParser;
use seekquarry\yioop\library\processors\PageProcessor;

/**
 * Class which may be used by TextProcessors to get a summary for a text
 * document that may later be used for indexing. It does this by doing
 * centroid-based clustering. It also generates a word cloud for a document
 * @author Mangesh Dahale mangeshadahale@gmail.com
 */
class CentroidSummarizer extends Summarizer
{
    /**
     * whether to output the results to the disk or not
     */
    const OUTPUT_TO_FILE = false;
    /**
     * The full disk location to save the result to
     */
    const OUTPUT_FILE_PATH = "/temp/centroid_summarizer_result.txt";
    /**
     * Generates a centroid with which every sentence is ranked with cosine
     * ranking method and also generates a word cloud.
     *
     * @param object $dom document object model of page to summarize
     * @param string $page complete raw page to generate the summary from.
     * @param string $lang language of the page to decide which stop words to
     *     call proper tokenizer.php of the specified language.
     * @return array array of summary and word cloud
     */
    public static function getSummary($dom, $page, $lang)
    {
        $page = self::pageProcessing($page);
        /* Format the document to remove characters other than periods and
           alphanumerics.
        */
        $formatted_doc = self::formatDoc($page);
        $out_sentences = self::getSentences($page);
        $stop_obj = PhraseParser::getTokenizer($lang);
        $sentences = self::removeStopWords($out_sentences, $stop_obj);
        /*  Splitting into terms */
        $terms = [];
        foreach ($sentences as $sentence) {
            $terms = array_merge($terms,
                PhraseParser::segmentSegment($sentence, $lang));
        }
        $n = count($out_sentences);
        $terms = array_filter($terms);
        list($word_cloud, $wc, $idf) =
            self::wordCloudAndCountsFromTermsSentences($terms, $sentences,
            $formatted_doc, $lang);
        if (strlen($page) < PageProcessor::$max_description_len
            || $n == 1) {
            //if input short only use above to get a word cloud
            $page = substr($page, 0,
                PageProcessor::$max_description_len);
            return [$page, $word_cloud];
        }
        ksort($wc);
        $t = count($terms);
        /* Calculate similarity measure between centroid and each sentence */
        $sim = [];
        for ($i = 0; $i < $n; $i++) {
            $a = 0;
            $b1 = 0;
            $b2 = 0;
            $c1 = 0;
            $c2 = 0;
            $d = 0;
            for ($k = 0; $k < $t; $k++) {
                    $wck = $wc[$k];
                    $idfk = $idf[$k];
                    $tmp = substr_count($sentences[$i], $terms[$k]);
                    $wik = ($tmp > 0) ? $idfk * (1 + log($tmp)) : 0;
                    $a += ($wik * $wck * $idfk);
                    $b1 += ($wik * $wik);
                    $c1 += ($wck * $wck);
            }
            $b2 = sqrt($b1);
            $c2 = sqrt($c1);
            $d = $b2 * $c2;
            if ($d == 0) {
                $sim[$i] = 0;
            } else {
                $sim[$i] = $a / $d;
            }
        }
        arsort($sim);
        /* Getting how many sentences should be there in summary */
        $top = self::numSentencesForSummary($out_sentences, $sim);
        $sum_array = [];
        $sum_array = array_keys(array_slice($sim, 0, $top - 1, true));
        sort($sum_array);
        $summary = '';
        $eos = ($lang == 'hi') ? "।" : "."; //default end of sentence symbol
        foreach ($sum_array as $key) {
            $compressed_sentence =
                PhraseParser::compressSentence($out_sentences[$key], $lang);
            $summary .= rtrim($compressed_sentence, $eos) . "$eos ";
        }
        if (self::OUTPUT_TO_FILE) {
            $output_file_contents = "";
            foreach ($sum_array as $key) {
                $compressed_sentence =
                    PhraseParser::compressSentence($out_sentences[$key],
                    $lang);
                $output_file_contents .= rtrim($compressed_sentence,
                    $eos) . "$eos\r\n";
            }
            file_put_contents(C\WORK_DIRECTORY . self::OUTPUT_FILE_PATH,
                $output_file_contents);
        }
        /* Summary of text summarization */
        return [$summary, $word_cloud];
    }
    /**
     * Calculates how many sentences to put in the summary to match the
     * MAX_DESCRIPTION_LEN.
     *
     * @param array $sentences sentences in doc in their original order
     * @param array $sim associative array of sentence-number-in-doc =>
     *      similarity score to centroid (sorted from highest to lowest score).
     * @return int number of sentences
     */
    public static function numSentencesForSummary($sentences, $sim)
    {
        $top = null;
        $count = 0;
        foreach ($sim as $key => $value)
        {
            if ($count < PageProcessor::$max_description_len) {
                $count += strlen($sentences[$key]);
                $top++;
            }
        }
        return $top;
    }
}
