<?php
/**
 * SeekQuarry/Yioop --
 * Open Source Pure PHP Search Engine, Crawler, and Indexer
 *
 * Copyright (C) 2009 - 2019  Chris Pollett chris@pollett.org
 *
 * LICENSE:
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * END LICENSE
 *
 * @author Charles Bocage charles.bocage@sjsu.edu
 * @license https://www.gnu.org/licenses/ GPL3
 * @link https://www.seekquarry.com/
 * @copyright 2009 - 2019
 * @filesource
 */
namespace seekquarry\yioop\library\summarizers;

use seekquarry\yioop\configs as C;
use seekquarry\yioop\library\PhraseParser;

/** For Yioop global defines used by subclasses*/
require_once __DIR__."/../../configs/Config.php";
/**
 * Base class for all summarizers that will hold common methods and
 * base functionality.
 *
 * @author Charles Bocage charles.bocage@sjsu.edu
 */
class Summarizer
{
    /**
     * Number of distinct terms to use in generating summary
     */
    const MAX_DISTINCT_TERMS = 1000;
    /**
     * Number of nonzero centroid components
     */
    const CENTROID_COMPONENTS = 50;
    /**
     * Number of words in word cloud
     */
    const WORD_CLOUD_LEN = 5;
    /**
     * The value to represent the weight for class one tags.
     */
    public static $CLASS_ONE_WEIGHT = 0;
    /**
     * The value to represent the weight for class two tags.
     */
    public static $CLASS_TWO_WEIGHT = 0;
    /**
     * The value to represent the weight for class three tags.
     */
    public static $CLASS_THREE_WEIGHT = 0;
    /**
     * The value to represent the weight for class four tags.
     */
    public static $CLASS_FOUR_WEIGHT = 0;
    /**
     * The value to represent the weight for class five tags.
     */
    public static $CLASS_FIVE_WEIGHT = 0;
    /**
     * Compute a summary of a document in a given language
     *
     * @param object $dom document object model used to locate items for
     *      summary
     * @param string $page raw document sentences should be extracted from
     * @param string $lang locale tag for language the summary is in
     */
    public static function getSummary($dom, $page, $lang) {
        throw \Exception("Not defined");
    }
    /**
     * This function will return additional weights if the term is in
     * certain tags of the document.  Each tag is grouped into 6 categories:
     * A
     * H1, H2
     * H3, H4, H5, H6
     * STRONG, B, EM, I, U, DL, OL, UL
     * Title
     * Plain Text = None of the above
     *
     * @param string $term the term to search for.
     * @param string $doc complete raw page to generate the summary from.
     *
     * @return int the additional weight of the term
     */
    public static function getAdditionalWeight($term, $doc)
    {
        $result = 0;
        $class_one_regex = "/(?s)<a.*?>(.*?" . $term . ".*?)<\/a>/iu";
        $class_two_regex = "/(?s)<h1.*?>(.*?" . $term . ".*?)<\/h1>|" .
            "<h2.*?>(.*?" . $term . ".*?)<\/h2>/iu";
        $class_three_regex = "/(?s)<h3.*?>(.*?" . $term . ".*?)<\/h3>|" .
            "<h4.*?>(.*?" . $term . ".*?)<\/h4>|" .
            "<h5.*?>(.*?" . $term . ".*?)<\/h5>|" .
            "<h6.*?>(.*?" . $term . ".*?)<\/h6>/iu";
        $class_four_regex = "/(?s)<strong.*?>(.*?" . $term . ".*?)<\/strong>|" .
            "<b.*?>(.*?" . $term . ".*?)<\/b>|" .
            "<em.*?>(.*?" . $term . ".*?)<\/em>|" .
            "<i.*?>(.*?" . $term . ".*?)<\/i>|" .
            "<u.*?>(.*?" . $term . ".*?)<\/u>|" .
            "<dl.*?>(.*?" . $term . ".*?)<\/dl>|" .
            "<ol.*?>(.*?" . $term . ".*?)<\/ol>|" .
            "<ul.*?>(.*?" . $term . ".*?)<\/ul>/iu";
        $class_five_regex = "/(?s)<title.*?>(.*?" . $term . ".*?)<\/title>/iu";
        $class_one_weight = self::$CLASS_ONE_WEIGHT;
        $class_two_weight = self::$CLASS_TWO_WEIGHT;
        $class_three_weight = self::$CLASS_THREE_WEIGHT;
        $class_four_weight = self::$CLASS_FOUR_WEIGHT;
        $class_five_weight = self::$CLASS_FIVE_WEIGHT;
        $class_one_matches = 0;
        $class_two_matches = 0;
        $class_three_matches = 0;
        $class_four_matches = 0;
        $class_five_matches = 0;
        $class_one_matches_count = 0;
        $class_two_matches_count = 0;
        $class_three_matches_count = 0;
        $class_four_matches_count = 0;
        $class_five_matches_count = 0;
        if (@strpos(mb_strtolower($doc),
            mb_strtolower($term), 0) !== false)
        {
            preg_match_all($class_one_regex, $doc, $class_one_matches);
            preg_match_all($class_two_regex, $doc, $class_two_matches);
            preg_match_all($class_three_regex, $doc, $class_three_matches);
            preg_match_all($class_four_regex, $doc, $class_four_matches);
            preg_match_all($class_five_regex, $doc, $class_five_matches);

            if (count($class_one_matches[0]) > 0) {
                for ($i = 1; $i < count($class_one_matches); $i++) {
                    $class_one_matches_count = $class_one_matches_count +
                        substr_count($class_one_matches[$i][0], $term);
                }
            }
            if (count($class_two_matches[0]) > 0) {
                for ($i = 1; $i < count($class_two_matches); $i++) {
                    $class_two_matches_count = $class_two_matches_count +
                        substr_count($class_two_matches[$i][0], $term);
                }
            }
            if (count($class_three_matches[0]) > 0) {
                for ($i = 1; $i < count($class_three_matches); $i++) {
                    $class_three_matches_count = $class_three_matches_count +
                        substr_count($class_three_matches[$i][0], $term);
                }
            }
            if (count($class_four_matches[0]) > 0) {
                for ($i = 1; $i < count($class_four_matches); $i++) {
                    $class_four_matches_count = $class_four_matches_count +
                        substr_count($class_four_matches[$i][0], $term);
                }
            }
            if (count($class_five_matches[0]) > 0) {
                for ($i = 1; $i < count($class_five_matches); $i++) {
                    $class_five_matches_count = $class_five_matches_count +
                        substr_count($class_five_matches[$i][0], $term);
                }
            }
            $result =
                ($class_one_weight * $class_one_matches_count) +
                ($class_two_weight * $class_two_matches_count) +
                ($class_three_weight * $class_three_matches_count) +
                ($class_four_weight * $class_four_matches_count) +
                ($class_five_weight * $class_five_matches_count);
        }
        return $result;
    }
    /**
     * Breaks any content into sentences by splitting it on spaces or carriage
     *   returns
     * @param string $content complete page.
     * @return array array of sentences from that content.
     */
    public static function getSentences($content)
    {
        $content = preg_replace([ "/\n+(\.| |\t)+/u",
            "/((\p{L}|\p{N}|\)|\}|\]){5,}\s?(\.|\|।|\!|\?|！|？|。))\s+/u",
            "/।/u", "/(\n|\r)(\n|\r)+/", "/।./u"], ["\n", "$1.\n", "।\n\n",
            "..\n", "।"], $content);
        $lines = preg_split('/\.\n/', $content, 0, PREG_SPLIT_NO_EMPTY);
        $lines = preg_replace("/\s+/", " ", $lines);
        return $lines;
    }
    /**
     * Formats the sentences to remove all characters except words,
     *   digits and spaces
     * @param string $sentence complete page.
     * @return string formatted sentences.
     */
    public static function formatSentence($sentence)
    {
        $sentence = trim(preg_replace('/[^\p{L}\p{N}\s]+/u',
            ' ', mb_strtolower($sentence)));
        return $sentence;
    }
    /**
     * Formats the document to remove carriage returns, hyphens and digits
     * as we will not be using digits in word cloud.
     * The formatted document generated by this function is only used to
     * compute centroid.
     * @param string $content formatted page.
     * @return string formatted document.
     */
    public static function formatDoc($content)
    {
        $substitute = ['/[\n\r\-]+/', '/[^\p{L}\s\.]+/u', '/\.+/'];
        $content = preg_replace($substitute, ' ', $content);
        return $content;
    }
    /**
     * This function does an additional processing on the page
     * such as removing all the tags from the page
     * @param string $page complete page.
     * @return string processed page.
     */
    public static function pageProcessing($page)
    {
        $substitutions = ['@<script[^>]*?>.*?</script>@si',
            '/\&nbsp\;|\&rdquo\;|\&ldquo\;|\&mdash\;/si',
            '@<style[^>]*?>.*?</style>@si', '/\t\n/', '/\s{2,}/'
        ];
        $page = preg_replace($substitutions, ' ', $page);
        $new_page = preg_replace("/\<br\s*(\/)?\s*\>/", "\n", $page);
        $changed = false;
        if ($new_page != $page) {
            $changed = true;
            $page = $new_page;
        }
        $page = preg_replace("/\<\/(h1|h2|h3|h4|h5|h6|table|tr|td|div|".
            "p|address|section)\s*\>/iu", "\n\n", $page);
        $page = preg_replace("/\<a/iu", " <a", $page);
        $page = html_entity_decode($page);
        $page = preg_replace("/\</", " <", $page);
        $page = strip_tags($page);
        if ($changed) {
            $page = preg_replace("/(\r?\n[\t| ]*){2}/", "\n", $page);
        }
        $page = preg_replace("/(\r?\n[\t| ]*)/", "\n", $page);
        $page = preg_replace("/\n\n\n+/", "\n\n", $page);
        return $page;
    }
    /**
     * Returns a new array of sentences without the stop words
     * @param array $sentences the array of sentences to process
     * @param object $stop_obj the class that has the stopworedRemover method
     * @return array a new array of sentences without the stop words
     */
    public static function removeStopWords($sentences, $stop_obj)
    {
        if ($stop_obj && method_exists($stop_obj, "stopwordsRemover")) {
            $results = $stop_obj->stopwordsRemover($sentences);
        } else {
            $results = $sentences;
        }
        return $results;
    }
    /**
     * Calculates an array with key terms and values their frequencies
     * based on a supplied sentence
     *
     * @param array $terms the list of all terms in the doc
     * @param array $sentence the sentences in the doc
     * @return array a two dimensional array where the word is the key and
     *      the frequency is the value
     */
    public static function getTermFrequencies($terms, $sentence)
    {
        $t = count($terms);
        $nk = array_fill(0, $t, 0);
        for ($j = 0; $j < $t; $j++) {
            $nk[$j] += preg_match_all("/\b" . preg_quote($terms[$j], '/') .
                "\b/iu", $sentence);
        }
        return array_combine($terms, $nk);
    }
    /**
     * Normalize the term frequency vector by dividing its entries by its
     * L_2 norm.
     * @param array $term_frequencies the array with the terms as the key
     *      and its frequency as the value
     * @return array array of term frequencies normalized
     */
    public static function normalizeTermFrequencies($term_frequencies)
    {
        $sum_of_squares = 0;
        $result_sum = 0;
        if (count($term_frequencies) == 0) {
            return [];
        }
        foreach ($term_frequencies as $k => $v) {
            $sum_of_squares += ($v * $v);
        }
        $square_root = sqrt($sum_of_squares);
        foreach ($term_frequencies as $k => $v) {
            if ($square_root == 0) {
                $result[$k] = 0;
            } else {
                $result[$k] = ($v / $square_root);
            }
        }
        return $result;
    }
    /**
     *
     */
    public static function wordCloudFromSummary($summary,  $lang,
        $term_frequencies = null)
    {
        if ($term_frequencies == null) {
            $stop_obj = PhraseParser::getTokenizer($lang);
            if ($stop_obj && method_exists($stop_obj, "stopwordsRemover")) {
                $summary = $stop_obj->stopwordsRemover($summary);
            }
            $summary = mb_strtolower($summary);
            $terms = PhraseParser::segmentSegment($summary, $lang);
            $term_frequencies = self::getTermFrequencies($terms, $summary);
        }
        arsort($term_frequencies);
        $top5 = array_slice($term_frequencies, 0 , 5);
        return array_keys($top5);
    }
    /**
     * @param array $terms
     * @param array $sentences
     * @param string $formatted_doc
     * @param string $lang
     */
    public static function wordCloudAndCountsFromTermsSentences(&$terms,
        $sentences, $formatted_doc, $lang)
    {
        $n = count($sentences);
        $terms_counts = array_count_values($terms);
        arsort($terms_counts);
        $terms_counts = array_slice($terms_counts, 0,
            self::MAX_DISTINCT_TERMS);
        $terms = array_unique(array_keys($terms_counts));
        $t = count($terms);
        if ($t == 0) {
            return ["", "", 0];
        }
        /* Initialize Nk [Number of sentences the term occurs] */
        $nk = [];
        $nk = array_fill(0, $t, 0);
        $nt = [];
        /* Count TF for each word */
        for ($i = 0; $i < $n; $i++) {
            for ($j = 0; $j < $t; $j++) {
                if (is_string($terms[$j]) &&
                    strpos($sentences[$i], $terms[$j]) !== false) {
                    $nk[$j]++;
                }
            }
        }
        /* Calculate weights of each term for every sentence */
        $w = [];
        $idf = [];
        $idf_temp = 0;
        for ($k = 0; $k < $t; $k++) {
            if ($nk[$k] == 0) {
                $idf_temp = 0;
                $tmp = 0;
            } else {
                $idf_temp = $n / $nk[$k];
                $tmp = log($idf_temp);
            }
            $idf[$k] = $tmp;
        }
        /* Count TF for finding centroid */
        $wc = [];
        $max_nt = -1;
        $b = "\b";
        if (in_array($lang, ["zh-CN", "ja", "ko"])) {
            $b = "";
        }
        set_error_handler(null);
        for ($j = 0; $j < $t; $j++) {
            $quoted = preg_quote($terms[$j]);
            $nt = @preg_match_all("/$b(" . $quoted . ")$b/ui", $formatted_doc,
                $matches); //$matches included for backwards compatibility
            $wc[$j] = $nt * $idf[$j];
            if (is_nan($wc[$j]) || is_infinite($wc[$j])) {
                $wc[$j] = 0;
            }
        }
        set_error_handler(C\NS_CONFIGS . "yioop_error_handler");
        /* Calculate centroid */
        arsort($wc);
        $centroid = array_slice($wc, 0, self::CENTROID_COMPONENTS, true);
        /* Initializing centroid weight array by 0 */
        $wc = array_fill(0, $t, 0);
        /* Word cloud */
        $i = 0;
        $word_cloud = [];
        foreach ($centroid as $key => $value) {
            $wc[$key] = $value;
            if ($i < self::WORD_CLOUD_LEN) {
                $word_cloud[$i] = $terms[$key];
            }
            $i++;
        }
        return [$word_cloud, $wc, $idf];
    }
}
